// Copyright 2017 by Yun Xiang, Tam Chantem, and Robert Dick.
// All rights reserved.

/*Copyright 2010, Yun Xiang<xiangyun@umich.edu>.*/
#include "RStd.h"
#include "Stats.h"
#include "Cont.h"

#include <iostream>
#include <iomanip>
#include <cmath>
#include <math.h>
#include <stdlib.h>
#include <ctime>
#include <cstdlib>

static const double PI = 3.14159;

using namespace std;

namespace rstd {
/*===========================================================================*/

double gamma_(double x)
{
    int i,k,m;
    double ga,gr,r = 1.0,z;

    static double g[] = {
        1.0,
        0.5772156649015329,
       -0.6558780715202538,
       -0.420026350340952e-1,
        0.1665386113822915,
       -0.421977345555443e-1,
       -0.9621971527877e-2,
        0.7218943246663e-2,
       -0.11651675918591e-2,
       -0.2152416741149e-3,
        0.1280502823882e-3,
       -0.201348547807e-4,
       -0.12504934821e-5,
        0.1133027232e-5,
       -0.2056338417e-6,
        0.6116095e-8,
        0.50020075e-8,
       -0.11812746e-8,
        0.1043427e-9,
        0.77823e-11,
       -0.36968e-11,
        0.51e-12,
       -0.206e-13,
       -0.54e-14,
        0.14e-14};

    if (x > 171.0) return 1e308;    // This value is an overflow flag.
    if (x == (int)x) {
        if (x > 0.0) {
            ga = 1.0;               // use factorial
            for (i=2;i<x;i++) {
               ga *= i;
            }
         }
         else
            ga = 1e308;
     }
     else {
        if (fabs(x) > 1.0) {
            z = fabs(x);
            m = (int)z;
            r = 1.0;
            for (k=1;k<=m;k++) {
                r *= (z-k);
            }
            z -= m;
        }
        else
            z = x;
        gr = g[24];
        for (k=23;k>=0;k--) {
            gr = gr*z+g[k];
        }
        ga = 1.0/(gr*z);
        if (fabs(x) > 1.0) {
            ga *= r;
            if (x < 0.0) {
                ga = -M_PI/(x*ga*sin(M_PI*x));
            }
        }
    }
    return ga;
    }


/*===========================================================================*/
 double erfcore( double x, int jint ) {	// Translation from Matlab lib
	double result = 0.0;

	// evaluate erf for |x| <= 0.46875
	double xbreak = 0.46578;
	int k = static_cast<int>(fabs(x) <= xbreak);
	if ( k != 0 ) {
		double a[5] = 
			{3.16112374387056560e00, 1.13864154151050156e02,
			 3.77485237685302021e02, 3.20937758913846947e03,
                       	 1.85777706184603153e-1};
		double b[4] = 
			{2.36012909523441209e01, 2.44024637934444173e02,
             		 1.28261652607737228e03, 2.84423683343917062e03};
		double y = fabs(x);
		double z = y*y;
		double xnum = a[4]*z;
		double xden = z;
		for ( int i = 0; i <=2; i++ ) {
			xnum = ( xnum + a[i] ) * z;
			xden = ( xden + b[i] ) * z;
		}
		result = x * ( xnum + a[3] ) / ( xden + b[3] );
		if ( jint != 0 )
			result = 1 - result;
		if ( jint == 2 )
			result = exp(z) * result;
	}

	// evaluate erfc for 0.46875 <= |x| <= 4.0
	k = ( fabs(x) > xbreak) && ( fabs(x) <= 4.0 );
	if ( k != 0 ) {
		double c[9] =
			{5.64188496988670089e-1, 8.88314979438837594e00,
			 6.61191906371416295e01, 2.98635138197400131e02,
			 8.81952221241769090e02, 1.71204761263407058e03,
			 2.05107837782607147e03, 1.23033935479799725e03,
			 2.15311535474403846e-8};
        	double d[8] = 
			{1.57449261107098347e01, 1.17693950891312499e02,
			 5.37181101862009858e02, 1.62138957456669019e03,
			 3.29079923573345963e03, 4.36261909014324716e03,
			 3.43936767414372164e03, 1.23033935480374942e03};
		double y = fabs(x);
		double xnum = c[8] * y;
		double xden = y;
		for ( int i = 0; i <= 6; i++) {
			xnum = ( xnum + c[i] ) * y;
			xden = ( xden + d[i] ) * y;
		}
		result = ( xnum + c[7] ) / ( xden + d[7] );
		if (jint != 2) {
			double z = static_cast<int>(y*16)/16;
			double del = (y-z)*(y+z);
			result = exp(-z*z)*exp(-del)*result;
		}
	}

	// evaluate erfc for |x| > 4.0
	k = ( fabs(x) > 4.0 );
	if ( k != 0 ) {
		double p[6] = 
			{3.05326634961232344e-1, 3.60344899949804439e-1,
			 1.25781726111229246e-1, 1.60837851487422766e-2,
			 6.58749161529837803e-4, 1.63153871373020978e-2};
        	double q[5] = 
			{2.56852019228982242e00, 1.87295284992346047e00,
			 5.27905102951428412e-1, 6.05183413124413191e-2,
			 2.33520497626869185e-3};
		double y = fabs(x);
		double z = 1/(y*y);
		double xnum = p[5] * z;
		double xden = z;
		for (int i = 0; i <= 3; i++) {
			xnum = ( xnum + p[i] ) * z;
			xden = ( xden + q[i] ) * z;
		}
		result = z * ( xnum + p[4] ) / ( xden + q[4] );
		result = ( 1/sqrt(PI) - result ) / y;
		
		if ( jint != 2 ) {
			z = static_cast<int>(y*16)/16;
			double del = ( y - z ) * ( y + z );
			result = exp(-z*z) * exp(-del) * result;
		}
	}

	// fix up for negative argument, erf, etc.
	if ( jint == 0 ) {
		k = ( x > xbreak );
		if ( k ) 
			result = ( 0.5 - result ) + 0.5;
		k = ( x < -xbreak );
		if ( k ) 
			result = ( -0.5 + result ) - 0.5;
	}
	else if ( jint == 1 ) {
		k = ( x < -xbreak );
		if ( k )
			result = 2.0 - result;
		
	}
	else {
		k = ( x < -xbreak );
		if ( k ) {
			double z = static_cast<int>(x*16)/16;
			double del = (x-z) * (x+z);
			double y = exp(z*z) * exp(del);
			result = (y*y) - result;
		}
	}
	return result;
}


double erfcinv( double y ) {	// Translation from Matlab lib
	double x = 0.0;
	// Coefficients in retional approximations
	double a[6] =  
		{ 1.370600482778535e-02, -3.051415712357203e-01,
		  1.524304069216834e+00, -3.057303267970988e+00,
		  2.710410832036097e+00, -8.862269264526915e-01 };
	double b[5] = 
		{ -5.319931523264068e-02,  6.311946752267222e-01,
		  -2.432796560310728e+00,  4.175081992982483e+00,
		  -3.320170388221430e+00 };
	double c[6] = 
		{ 5.504751339936943e-03, 2.279687217114118e-01, 
		  1.697592457770869e+00, 1.802933168781950e+00, 
		  -3.093354679843504e+00, -2.077595676404383e+00 };
	double d[4] = 
		{ 7.784695709041462e-03, 3.224671290700398e-01,
		  2.445134137142996e+00, 3.754408661907416e+00 };

	// Define break-points
	double ylow = 0.0485;
	double yhigh = 1.9515;

	// Rational approximation for central region
	int k = ( ylow <= y ) & ( y <= yhigh );
	double q, r;
	if (k) {
		q = y - 1;
		r = q * q;
		x = (((((a[0]*r+a[1])*r+a[2])*r+a[3])*r+a[4])*r+a[5])*q /
		    (((((b[0]*r+b[1])*r+b[2])*r+b[3])*r+b[4])*r+1);
	}
	
	// Rational approximation for lower region
	k = ( 0 <= y ) & ( y <= ylow );
	if (k) {
		q = sqrt(-2*log(y/2));
		x = -((((c[0]*q+c[1])*q+c[2])*q+c[3])*q+c[4])*q+c[5] /
		     ((((d[0]*q+d[1])*q+d[2])*q+d[3])*q+1);
	}

	// Rational approximation for upper region
	k = ( yhigh < y ) & ( y < 2 );
	if ( k ) {
		q = sqrt(-2*log(1-y/2));
		x = -(((((c[0]*q+c[1])*q+c[2])*q+c[3])*q+c[4])*q+c[5]) / 
		      ((((d[0]*q+d[1])*q+d[2])*q+d[3])*q+1);
	}

	// The relative error of the approximation has absolute value less
	// than 1.13e-9. One iteration of Halley's rational method ( third
	// order ) gives full machine precision.

	// Newton's method: new x = x - f/f'
	// Halley's method: new x = x - 1/(f'/f - (f"/f')/2)
	// This function: f = erfc(x) - y, f' = -2/sqrt(pi)*exp(-x^2), f' =
	// -2*x*f'  
	
	// Newton's correction
	double u = (erfc(x) - y) / (-2/sqrt(PI) * exp(pow(-x, 2)));

	// Halley's step
	x = x - u / ( 1 + x * u );
	return x;
	
}


/*===========================================================================*/
double erfc(double x) {		// Translation from Matlab lib
	return erfcore( x, 1 );
}

/*===========================================================================*/
double erf(double x) {		// Translation from Matlab lib
	return erfcore( x, 0 );
}

/*===========================================================================*/
  double unirnd(){
    return rand()/(RAND_MAX+1.0);
  }
/*===========================================================================*/
  double wblrnd(double eta, double beta){
    if (beta != 0)
      return wblinv(unirnd(), eta, beta);
    return eta;
  }
/*===========================================================================*/
  double normrnd(double mu, double sigma){
    if (mu != 0)
      {
	double a = norminv(unirnd(), mu, sigma);
	if (a < 0) a = 0;
	if (a > mu+6*sigma) a = mu+6*sigma;
	else if (a < mu-6*sigma) a = mu-6*sigma;
	return a;
      }
    return mu;
  }

/*===========================================================================*/
  //considering the process variation here!!
  double wblinv_PV(double p,double eta, double mu, double sigma, int n){
    if (p > (1/INF)){
      if (sigma < (1/INF)){
				double x = pow(-log(1-p)/n,1/mu)*eta;
				return x;
      }
      else{
				double b = mu*mu-2*sigma*sigma*log(-log(1-p)/n);
				if (b<0){
	  			//cout<<"solution has negative value under square root!! while input p is: "<<p<<endl;
	  			double x = pow(-log(1-p)/n,1/mu)*eta;
	  			return x;
				}
				else{
	  			double a = -(-mu+sqrt(b))/pow(sigma,2);
	  			double x = eta*exp(a);
	  			return x;
				}
      }
    }
    return 0;
  }
  /*===========================================================================*/     
  long double wblcdf_PV(double x, double eta, double mu, double sigma, int n) {
    long double z;
    long double p;
    //Rdump(sigma);
    //Rdump(x);
    //Rdump(eta);
    //Rdump(x/eta);
    //Rdump(log(x/eta));
    //Rdump(sigma * sigma * log(x/eta) / 2);
    double a = sigma*sigma*log(x/eta)/2;
    //Rdump(a);
    //Rdump(x/eta);
    //Rdump(mu-a);
    //Rdump(pow(x / eta, mu - a));
    //Rdump(n * pow(x / eta, mu - a));
    z = n*pow(x/eta, mu-a);
    //Rdump(z);
    //Rdump(exp(-z));
    p=1-exp(-z);
    //Rdump(p);

    return p;
  }

/*===========================================================================*/
double normcdf( double x, double mu, double sigma) {// Translation from Matlab lib
	double p;
	double z;
	z = (x-mu)/sigma;
	p = 0.5 * erfc(-z / sqrt(2));
	return p;
}
/*===========================================================================*/
double norminv( double p, double mu, double sigma) {	
	double x0 = -sqrt(2) * erfcinv(2*p);
	double x = sigma * x0 + mu;
	return x;
}

/*===========================================================================*/
double logncdf( double x, double mu, double sigma) {// Translation from Matlab lib
	double p;
	double z;
	z = (log(x)-mu) / sigma;
	p = 0.5 * erfc(-z/sqrt(2));
	return p;
}

  /*========================================================================*/
  double wblcdf (double x, double mu, double sigma)
  {
    double z;
    double p;
    z = pow(x/mu, sigma);
    p=1-exp(-z);
    return p;
  }
  /*========================================================================*/
  double wblpdf (double x, double mu, double sigma)
  {
    double z;
    double p;
    z = exp(-pow(x/mu, sigma));
    p=z*(sigma/mu)*pow(x/mu,sigma-1);
    return p;
  }

  /*========================================================================*/
  double wblinv(double p, double mu, double sigma)
  {
    double x= pow(-log(1-p),1/sigma)*mu;
    return x;
  }


/*===========================================================================*/
double logninv( double p, double mu, double sigma) {// Translation from Matlab lib
	double logx0 = -sqrt(2) * erfcinv(2*p);
	double x = exp( sigma * logx0 + mu );
	return x;
}



/*===========================================================================*/
double lognpdf( double x, double mu, double sigma) {
	double y;
	y = exp(-0.5 * pow(((log(x) - mu) / sigma), 2)) / (x*sqrt(2*PI) *
		sigma);
	return y;
} 

/*===========================================================================*/
double lognmean(double mu, double sigma) {
	return (exp(mu+sigma*sigma/2));
}

/*===========================================================================*/
void lognfit(const RVector<double>& data, double& mu, double& sigma) {
RVector<double> x(data.size());
	for (int i = 0; i < data.size(); i++) {
		x[i] = log(data[i]);
	}
	normfit(x, mu, sigma);
}

/*===========================================================================*/
void lognfit(const RVector<double> & Ft, const RVector<double>& t, 
double & mu, double& sigma) {
	Rassert(Ft.size() == t.size());
	int lsize = Ft.size();
	double order1 = 0.0;
	double order2 = 0.0;
	double midt;
	for (int i = 0; i < lsize-1; i++) {
		midt = (t[i+1]+t[i])/2;
		order1 += (Ft[i+1]-Ft[i])*midt;
		order2 += (Ft[i+1]-Ft[i])*midt*midt;
	}
//	cout << "order1: " << order1 << endl;
//	cout << "order2: " << order2 << endl;
	mu = log(pow(order1, 4)/order2)/2;
	sigma = sqrt(log(order2/(order1*order1)));
}
  /*===================================================================*/
void wblfit(const RVector<double> & Ft, const RVector<double>& t, 
double & mu, double& sigma) {
	Rassert(Ft.size() == t.size());
	int lsize = Ft.size();
	double order1 = 0.0;
	double order2 = 0.0;
	double midt;
	int i;
	for (i = 0; i < lsize-1; i++) {
		midt = (t[i+1]+t[i])/2;
		order1 += (Ft[i+1]-Ft[i])*midt;
		order2 += (Ft[i+1]-Ft[i])*midt*midt;
	}
	//cout << "order1: " << order1 << endl;
	//cout << "order2: " << order2 << endl;
	double soln = (order1*order1)/order2;
        double diff = 10000;
	double diff0;
	double a=1.0;
	double step = 0.01;
	sigma = a;
	for (i = 0; i < 1000;i++)
	  {
	    a += step;
	    diff0 = fabs(soln - gamma(1+(1/a))*gamma(1+(1/a))/gamma(1+(2/a)));
	    if (diff0 < diff)
	      {
		diff = diff0;
		sigma = a;
	      }
	  }
	mu = order1/rstd::gamma_(1+1/sigma);
	
	//	mu = log(pow(order1, 4)/order2)/2;
	//sigma = sqrt(log(order2/(order1*order1)));
}
  /*========================================================================*/

void wblfit(const RVector<double> & mu_vec, const RVector<double> & sigma_vec,
const double & threshold, double & mu, double & sigma) {
	Rassert(mu_vec.size() == sigma_vec.size());
	RVector<double> t;
	RVector<double> Ft;
	const double LTIME = 0.0;
	const double TIME_INTERVAL = 0.01;
	int i = 0;
	while ( 1 ) {
		t.push_back(LTIME+TIME_INTERVAL*i);
		double tmp_Ft = 1.0;
		for (int j = 0; j < mu_vec.size(); j++) {
			tmp_Ft *= (1-wblcdf(t[i], mu_vec[j], sigma_vec[j]));
		}
		tmp_Ft =1 - tmp_Ft;
		Ft.push_back(tmp_Ft);
		if (tmp_Ft > threshold)
			break;
		i++;
	}
	//cout << "size of Ft: " << Ft.size() << endl;
	
	wblfit(Ft, t, mu, sigma);

}


/*==========================================================================*/	
void lognfit(const RVector<double> & mu_vec, const RVector<double> & sigma_vec,
const double & threshold, double & mu, double & sigma) {
	Rassert(mu_vec.size() == sigma_vec.size());
	RVector<double> t;
	RVector<double> Ft;
	const double LTIME = 0.0;
	const double TIME_INTERVAL = 0.01;
	int i = 0;
	while ( 1 ) {
		t.push_back(LTIME+TIME_INTERVAL*i);
		double tmp_Ft = 1.0;
		for (int j = 0; j < mu_vec.size(); j++) {
			tmp_Ft *= (1-logncdf(t[i], mu_vec[j], sigma_vec[j]));
		}
		tmp_Ft =1 - tmp_Ft;
		Ft.push_back(tmp_Ft);
		if (tmp_Ft > threshold)
			break;
		i++;
	}
//	cout << "size of Ft: " << Ft.size() << endl;
	
	lognfit(Ft, t, mu, sigma);
}

/*===========================================================================*/
void normfit(const RVector<double> & x, double & mu, double & sigma) {
	double sumx = 0.0;
	for (int i = 0; i < x.size(); i++) 
		sumx += x[i];
	int n = x.size();
	
	double muhat = sumx/n;
	mu = muhat;
	if (n > 1) {
		RVector<double> xc(x.size());
		for (int i = 0; i < x.size(); i++) 
			xc[i] = x[i] - muhat;
		double sumxc2 = 0.0;
		for (int i = 0; i < xc.size(); i++)
			sumxc2 += xc[i] * xc[i];
		sigma = sqrt(sumxc2/(n-1));
	} else
		sigma = 0.0;
}
/*===========================================================================*/
  double wblintegral(double mu, double sigma, double start, double end)
  {
    double window_size = 0.001;
    double step, step_size;
    double area = 0;
    step_size=(end-start)*window_size;
    step = start;
    int i = 0;
    for (i = 0; i < 1/window_size; i++){     
      area+=(wblpdf(step,mu,sigma)+wblpdf(step+step_size,mu,sigma))/2;
      step=step+step_size;
    }
    return area*step_size;
  }

/*===========================================================================*/
void lognfit2_test() {
	
	RVector<double> mu_v;
	mu_v.push_back(2.17759);
	mu_v.push_back(2.17759);
	mu_v.push_back(2.17759);
	mu_v.push_back(2.17759);
	RVector<double> sigma_v;
	sigma_v.push_back(5);
	sigma_v.push_back(5);
	sigma_v.push_back(5);
	sigma_v.push_back(5);

	double mu, sigma;
	wblfit(mu_v, sigma_v, 0.999, mu, sigma);
	//cout << "mu: " << mu << endl;
	//cout << "sigma: " << sigma << endl;

}

}
